/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 3.0.12
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */

#define SWIG_PYTHON_STRICT_BYTE_CHAR
#if defined(_MSC_VER)
/* Prevent "non-constant aggregate initializer" errors from Python.h in
 * Python 3.9 */
# pragma warning(default : 4204)
#endif



#ifndef SWIGPYTHON
#define SWIGPYTHON
#endif

#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif


#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03000000
  char *cstr;
  char *newstr;
  Py_ssize_t len;
  str = PyUnicode_AsUTF8String(str);
  PyBytes_AsStringAndSize(str, &cstr, &len);
  newstr = (char *) malloc(len+1);
  memcpy(newstr, cstr, len+1);
  Py_XDECREF(str);
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#else
#  define SWIG_Python_str_DelForPy3(x) 
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

/* Add PyOS_snprintf for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# if defined(_MSC_VER) || defined(__BORLANDC__) || defined(_WATCOM)
#  define PyOS_snprintf _snprintf
# else
#  define PyOS_snprintf snprintf
# endif
#endif

/* A crude PyString_FromFormat implementation for old Pythons */
#if PY_VERSION_HEX < 0x02020000

#ifndef SWIG_PYBUFFER_SIZE
# define SWIG_PYBUFFER_SIZE 1024
#endif

static PyObject *
PyString_FromFormat(const char *fmt, ...) {
  va_list ap;
  char buf[SWIG_PYBUFFER_SIZE * 2];
  int res;
  va_start(ap, fmt);
  res = vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);
  return (res < 0 || res >= (int)sizeof(buf)) ? 0 : PyString_FromString(buf);
}
#endif

#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

/* A crude PyExc_StopIteration exception for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# ifndef PyExc_StopIteration
#  define PyExc_StopIteration PyExc_RuntimeError
# endif
# ifndef PyObject_GenericGetAttr
#  define PyObject_GenericGetAttr 0
# endif
#endif

/* Py_NotImplemented is defined in 2.1 and up. */
#if PY_VERSION_HEX < 0x02010000
# ifndef Py_NotImplemented
#  define Py_NotImplemented PyExc_RuntimeError
# endif
#endif

/* A crude PyString_AsStringAndSize implementation for old Pythons */
#if PY_VERSION_HEX < 0x02010000
# ifndef PyString_AsStringAndSize
#  define PyString_AsStringAndSize(obj, s, len) {*s = PyString_AsString(obj); *len = *s ? strlen(*s) : 0;}
# endif
#endif

/* PySequence_Size for old Pythons */
#if PY_VERSION_HEX < 0x02000000
# ifndef PySequence_Size
#  define PySequence_Size PySequence_Length
# endif
#endif

/* PyBool_FromLong for old Pythons */
#if PY_VERSION_HEX < 0x02030000
static
PyObject *PyBool_FromLong(long ok)
{
  PyObject *result = ok ? Py_True : Py_False;
  Py_INCREF(result);
  return result;
}
#endif

/* Py_ssize_t for old Pythons */
/* This code is as recommended by: */
/* http://www.python.org/dev/peps/pep-0353/#conversion-guidelines */
#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
# define PY_SSIZE_T_MAX INT_MAX
# define PY_SSIZE_T_MIN INT_MIN
typedef inquiry lenfunc;
typedef intargfunc ssizeargfunc;
typedef intintargfunc ssizessizeargfunc;
typedef intobjargproc ssizeobjargproc;
typedef intintobjargproc ssizessizeobjargproc;
typedef getreadbufferproc readbufferproc;
typedef getwritebufferproc writebufferproc;
typedef getsegcountproc segcountproc;
typedef getcharbufferproc charbufferproc;
static long PyNumber_AsSsize_t (PyObject *x, void *SWIGUNUSEDPARM(exc))
{
  long result = 0;
  PyObject *i = PyNumber_Int(x);
  if (i) {
    result = PyInt_AsLong(i);
    Py_DECREF(i);
  }
  return result;
}
#endif

#if PY_VERSION_HEX < 0x02050000
#define PyInt_FromSize_t(x) PyInt_FromLong((long)x)
#endif

#if PY_VERSION_HEX < 0x02040000
#define Py_VISIT(op)				\
  do { 						\
    if (op) {					\
      int vret = visit((op), arg);		\
      if (vret)					\
        return vret;				\
    }						\
  } while (0)
#endif

#if PY_VERSION_HEX < 0x02030000
typedef struct {
  PyTypeObject type;
  PyNumberMethods as_number;
  PyMappingMethods as_mapping;
  PySequenceMethods as_sequence;
  PyBufferProcs as_buffer;
  PyObject *name, *slots;
} PyHeapTypeObject;
#endif

#if PY_VERSION_HEX < 0x02030000
typedef destructor freefunc;
#endif

#if ((PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION > 6) || \
     (PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION > 0) || \
     (PY_MAJOR_VERSION > 3))
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)
#endif

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred()) PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    char *tmp;
    PyObject *old_str = PyObject_Str(value);
    PyErr_Clear();
    Py_XINCREF(type);

    PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    if (PY_VERSION_HEX >= 0x02030000) /* For 2.3 or later, use the PyGILState calls */
#      define SWIG_PYTHON_USE_GIL
#    endif
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;


/* -----------------------------------------------------------------------------
 * Wrapper of PyInstanceMethod_New() used in Python 3
 * It is exported to the generated module, used for -fastproxy
 * ----------------------------------------------------------------------------- */
#if PY_VERSION_HEX >= 0x03000000
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func)
{
  return PyInstanceMethod_New(func);
}
#else
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *SWIGUNUSEDPARM(func))
{
  return NULL;
}
#endif

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
#if !defined(SWIG_PYTHON_OUTPUT_TUPLE)
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
#else
  PyObject*   o2;
  PyObject*   o3;
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyTuple_Check(result)) {
      o2 = result;
      result = PyTuple_New(1);
      PyTuple_SET_ITEM(result, 0, o2);
    }
    o3 = PyTuple_New(1);
    PyTuple_SET_ITEM(o3, 0, obj);
    o2 = result;
    result = PySequence_Concat(o2, o3);
    Py_DECREF(o2);
    Py_DECREF(o3);
  }
  return result;
#endif
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

/* A functor is a function object with one single object argument */
#if PY_VERSION_HEX >= 0x02020000
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);
#else
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunction(functor, "O", obj);
#endif

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/*  How to access Py_None */
#if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#  ifndef SWIG_PYTHON_NO_BUILD_NONE
#    ifndef SWIG_PYTHON_BUILD_NONE
#      define SWIG_PYTHON_BUILD_NONE
#    endif
#  endif
#endif

#ifdef SWIG_PYTHON_BUILD_NONE
#  ifdef Py_None
#   undef Py_None
#   define Py_None SWIG_Py_None()
#  endif
SWIGRUNTIMEINLINE PyObject * 
_SWIG_Py_None(void)
{
  PyObject *none = Py_BuildValue((char*)"");
  Py_DECREF(none);
  return none;
}
SWIGRUNTIME PyObject * 
SWIG_Py_None(void)
{
  static PyObject *SWIG_STATIC_POINTER(none) = _SWIG_Py_None();
  return none;
}
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  return data ? data->implicitconv : 0;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
#if (PY_VERSION_HEX < 0x02020000)
      data->newraw = 0;
#else
      data->newraw = PyObject_GetAttrString(data->klass, (char *)"__new__");
#endif
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, (char *)"__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
#ifdef METH_O
      data->delargs = !(flags & (METH_O));
#else
      data->delargs = 0;
#endif
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_INCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
#ifdef METH_NOARGS
SwigPyObject_repr(SwigPyObject *v)
#else
SwigPyObject_repr(SwigPyObject *v, PyObject *args)
#endif
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
# ifdef METH_NOARGS
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# else
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next, args);
# endif
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *val = NULL, *type = NULL, *tb = NULL;
      PyErr_Fetch(&val, &type, &tb);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        res = SWIG_Python_CallFunctor(destroy, tmp);
        Py_DECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(val, type, tb);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
#ifndef METH_O
  PyObject *tmp = 0;
  if (!PyArg_ParseTuple(next,(char *)"O:append", &tmp)) return NULL;
  next = tmp;
#endif
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
#ifdef METH_NOARGS
SwigPyObject_next(PyObject* v)
#else
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_disown(PyObject *v)
#else
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_acquire(PyObject *v)
#else
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
#if (PY_VERSION_HEX < 0x02020000)
  if (!PyArg_ParseTuple(args,(char *)"|O:own",&val))
#elif (PY_VERSION_HEX < 0x02050000)
  if (!PyArg_UnpackTuple(args, (char *)"own", 0, 1, &val)) 
#else
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) 
#endif
    {
      return NULL;
    } 
  else
    {
      SwigPyObject *sobj = (SwigPyObject *)v;
      PyObject *obj = PyBool_FromLong(sobj->own);
      if (val) {
#ifdef METH_NOARGS
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v);
	} else {
	  SwigPyObject_disown(v);
	}
#else
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v,args);
	} else {
	  SwigPyObject_disown(v,args);
	}
#endif
      } 
      return obj;
    }
}

#ifdef METH_O
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_NOARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_NOARGS,  (char *)"acquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS, (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_O,       (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_NOARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,    METH_NOARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#else
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_VARARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_VARARGS,  (char *)"acquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS,  (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_VARARGS,  (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_VARARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,   METH_VARARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#endif

#if PY_VERSION_HEX < 0x02020000
SWIGINTERN PyObject *
SwigPyObject_getattr(SwigPyObject *sobj,char *name)
{
  return Py_FindMethod(swigobject_methods, (PyObject *)sobj, name);
}
#endif

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#elif PY_VERSION_HEX >= 0x02050000 /* 2.5.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#elif PY_VERSION_HEX >= 0x02020000 /* 2.2.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_true_divide */
#elif PY_VERSION_HEX >= 0x02000000 /* 2.0.0 */
    0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_or */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyObject",               /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
#if PY_VERSION_HEX < 0x02020000
      (getattrfunc)SwigPyObject_getattr,    /* tp_getattr */
#else
      (getattrfunc)0,                       /* tp_getattr */
#endif
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
      0,                                    /* tp_prev */
#endif
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpyobject_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
#endif
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME int
SwigPyPacked_print(SwigPyPacked *v, FILE *fp, int SWIGUNUSEDPARM(flags))
{
  char result[SWIG_BUFFER_SIZE];
  fputs("<Swig Packed ", fp); 
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    fputs("at ", fp); 
    fputs(result, fp); 
  }
  fputs(v->ty->name,fp); 
  fputs(">", fp);
  return 0; 
}
  
SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((char *)v->pack, (char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyPacked",               /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      (printfunc)SwigPyPacked_print,        /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
      0,                                    /* tp_prev */
#endif
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpypacked_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
#endif
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIMEINLINE PyObject *
_SWIG_This(void)
{
    return SWIG_Python_str_FromChar("this");
}

static PyObject *swig_this = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (swig_this == NULL)
    swig_this = _SWIG_This();
  return swig_this;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if (!defined(SWIG_PYTHON_SLOW_GETSET_THIS) && (PY_VERSION_HEX >= 0x02030000))
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
    }
    if (!SWIG_IsOK(res) && obj == Py_None) {
      if (ptr)
        *ptr = 0;
      if (PyErr_Occurred())
        PyErr_Clear();
      res = SWIG_OK;
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    
    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc) 
      return SWIG_ERROR;
    if (ty) {
      swig_cast_info *tc = SWIG_TypeCheck(desc,ty);
      if (tc) {
        int newmemory = 0;
        *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
        assert(!newmemory); /* newmemory handling not yet implemented */
      } else {
        return SWIG_ERROR;
      }
    } else {
      *ptr = vptr;
    }
    return SWIG_OK;
  }
}

/* Convert a packed value value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
#if (PY_VERSION_HEX >= 0x02020000)
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      PyObject *key = SWIG_This();
      PyObject_SetAttr(inst, key, swig_this);
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    inst = ((PyTypeObject*) data->newargs)->tp_new((PyTypeObject*) data->newargs, Py_None, Py_None);
    if (inst) {
      PyObject_SetAttr(inst, SWIG_This(), swig_this);
      Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
#else
#if (PY_VERSION_HEX >= 0x02010000)
  PyObject *inst = 0;
  PyObject *dict = PyDict_New();
  if (dict) {
    PyDict_SetItem(dict, SWIG_This(), swig_this);
    inst = PyInstance_NewRaw(data->newargs, dict);
    Py_DECREF(dict);
  }
  return (PyObject *) inst;
#else
  PyInstanceObject *inst = PyObject_NEW(PyInstanceObject, &PyInstance_Type);
  if (inst == NULL) {
    return NULL;
  }
  inst->in_class = (PyClassObject *)data->newargs;
  Py_INCREF(inst->in_class);
  inst->in_dict = PyDict_New();
  if (inst->in_dict == NULL) {
    Py_DECREF(inst);
    return NULL;
  }
#ifdef Py_TPFLAGS_HAVE_WEAKREFS
  inst->in_weakreflist = NULL;
#endif
#ifdef Py_TPFLAGS_GC
  PyObject_GC_Init(inst);
#endif
  PyDict_SetItem(inst->in_dict, SWIG_This(), swig_this);
  return (PyObject *) inst;
#endif
#endif
}

SWIGRUNTIME void
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
 PyObject *dict;
#if (PY_VERSION_HEX >= 0x02020000) && !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
 PyObject **dictptr = _PyObject_GetDictPtr(inst);
 if (dictptr != NULL) {
   dict = *dictptr;
   if (dict == NULL) {
     dict = PyDict_New();
     *dictptr = dict;
   }
   PyDict_SetItem(dict, SWIG_This(), swig_this);
   return;
 }
#endif
 dict = PyObject_GetAttrString(inst, (char*)"__dict__");
 PyDict_SetItem(dict, SWIG_This(), swig_this);
 Py_DECREF(dict);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      SWIG_Python_SetSwigThis(obj[0], obj[1]);
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
# ifdef SWIGPY_USE_CAPSULE
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
# else
    type_pointer = PyCObject_Import((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION,
				    (char*)"type_pointer" SWIG_TYPE_TABLE_NAME);
# endif
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

#if PY_MAJOR_VERSION < 2
/* PyModule_AddObject function was introduced in Python 2.0.  The following function
   is copied out of Python/modsupport.c in python version 2.3.4 */
SWIGINTERN int
PyModule_AddObject(PyObject *m, char *name, PyObject *o)
{
  PyObject *dict;
  if (!PyModule_Check(m)) {
    PyErr_SetString(PyExc_TypeError, "PyModule_AddObject() needs module as first arg");
    return SWIG_ERROR;
  }
  if (!o) {
    PyErr_SetString(PyExc_TypeError, "PyModule_AddObject() needs non-NULL value");
    return SWIG_ERROR;
  }
  
  dict = PyModule_GetDict(m);
  if (dict == NULL) {
    /* Internal error -- modules must have a dict! */
    PyErr_Format(PyExc_SystemError, "module '%s' has no __dict__",
		 PyModule_GetName(m));
    return SWIG_ERROR;
  }
  if (PyDict_SetItemString(dict, name, o))
    return SWIG_ERROR;
  Py_DECREF(o);
  return SWIG_OK;
}
#endif

SWIGRUNTIME void
#ifdef SWIGPY_USE_CAPSULE
SWIG_Python_DestroyModule(PyObject *obj)
#else
SWIG_Python_DestroyModule(void *vptr)
#endif
{
#ifdef SWIGPY_USE_CAPSULE
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
#else
  swig_module_info *swig_module = (swig_module_info *) vptr;
#endif
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  swig_this = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
#ifdef SWIGPY_USE_CAPSULE
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#else
  PyObject *pointer = PyCObject_FromVoidPtr((void *) swig_module, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#endif
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
#ifdef SWIGPY_USE_CAPSULE
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
#else
    descriptor = (swig_type_info *) PyCObject_AsVoidPtr(obj);
#endif
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
#ifdef SWIGPY_USE_CAPSULE
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
#else
      obj = PyCObject_FromVoidPtr(descriptor, NULL);
#endif
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      char *tmp;
      PyObject *old_str = PyObject_Str(value);
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, tmp = SWIG_Python_str_AsChar(old_str));
      } else {
	PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_apr_array_header_t swig_types[0]
#define SWIGTYPE_p_apr_file_t swig_types[1]
#define SWIGTYPE_p_apr_hash_t swig_types[2]
#define SWIGTYPE_p_apr_int32_t swig_types[3]
#define SWIGTYPE_p_apr_int64_t swig_types[4]
#define SWIGTYPE_p_apr_pool_t swig_types[5]
#define SWIGTYPE_p_apr_size_t swig_types[6]
#define SWIGTYPE_p_char swig_types[7]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[8]
#define SWIGTYPE_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[9]
#define SWIGTYPE_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[10]
#define SWIGTYPE_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[11]
#define SWIGTYPE_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[12]
#define SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t swig_types[13]
#define SWIGTYPE_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t swig_types[14]
#define SWIGTYPE_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t swig_types[15]
#define SWIGTYPE_p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[16]
#define SWIGTYPE_p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[17]
#define SWIGTYPE_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t swig_types[18]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_void__p_unsigned_char swig_types[20]
#define SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[22]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[26]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t swig_types[27]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_int swig_types[29]
#define SWIGTYPE_p_long swig_types[30]
#define SWIGTYPE_p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[31]
#define SWIGTYPE_p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[32]
#define SWIGTYPE_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[33]
#define SWIGTYPE_p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[34]
#define SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t swig_types[35]
#define SWIGTYPE_p_p_f_p_void__p_unsigned_char swig_types[36]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[37]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_p_svn_checksum_t swig_types[39]
#define SWIGTYPE_p_p_svn_delta_editor_t swig_types[40]
#define SWIGTYPE_p_p_svn_delta_path_driver_state_t swig_types[41]
#define SWIGTYPE_p_p_svn_txdelta_op_t swig_types[42]
#define SWIGTYPE_p_p_svn_txdelta_stream_t swig_types[43]
#define SWIGTYPE_p_p_svn_txdelta_window_t swig_types[44]
#define SWIGTYPE_p_p_void swig_types[45]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[46]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[47]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[48]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[49]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[50]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[51]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[52]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[53]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[54]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[55]
#define SWIGTYPE_p_svn_cache_config_t swig_types[56]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[57]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[58]
#define SWIGTYPE_p_svn_checksum_t swig_types[59]
#define SWIGTYPE_p_svn_client__shelf_version_t swig_types[60]
#define SWIGTYPE_p_svn_client_conflict_option_t swig_types[61]
#define SWIGTYPE_p_svn_commit_info_t swig_types[62]
#define SWIGTYPE_p_svn_config_t swig_types[63]
#define SWIGTYPE_p_svn_delta_editor_t swig_types[64]
#define SWIGTYPE_p_svn_delta_path_driver_state_t swig_types[65]
#define SWIGTYPE_p_svn_depth_t swig_types[66]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[67]
#define SWIGTYPE_p_svn_dirent_t swig_types[68]
#define SWIGTYPE_p_svn_errno_t swig_types[69]
#define SWIGTYPE_p_svn_error_t swig_types[70]
#define SWIGTYPE_p_svn_fs_lock_target_t swig_types[71]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[72]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[73]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[74]
#define SWIGTYPE_p_svn_location_segment_t swig_types[75]
#define SWIGTYPE_p_svn_lock_t swig_types[76]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[77]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[78]
#define SWIGTYPE_p_svn_log_entry_t swig_types[79]
#define SWIGTYPE_p_svn_merge_range_t swig_types[80]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[81]
#define SWIGTYPE_p_svn_node_kind_t swig_types[82]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[83]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[84]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[85]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[86]
#define SWIGTYPE_p_svn_opt_subcommand_desc3_t swig_types[87]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[88]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[89]
#define SWIGTYPE_p_svn_prop_kind swig_types[90]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[91]
#define SWIGTYPE_p_svn_stream_t swig_types[92]
#define SWIGTYPE_p_svn_string_t swig_types[93]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[94]
#define SWIGTYPE_p_svn_tristate_t swig_types[95]
#define SWIGTYPE_p_svn_txdelta_op_t swig_types[96]
#define SWIGTYPE_p_svn_txdelta_stream_t swig_types[97]
#define SWIGTYPE_p_svn_txdelta_window_t swig_types[98]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[99]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[100]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[101]
#define SWIGTYPE_p_svn_version_extended_t swig_types[102]
#define SWIGTYPE_p_svn_version_t swig_types[103]
#define SWIGTYPE_p_svn_wc_external_item2_t swig_types[104]
#define SWIGTYPE_p_unsigned_char swig_types[105]
#define SWIGTYPE_p_unsigned_long swig_types[106]
#define SWIGTYPE_p_void swig_types[107]
static swig_type_info *swig_types[109];
static swig_module_info swig_module = {swig_types, 108, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#if (PY_VERSION_HEX <= 0x02000000)
# if !defined(SWIG_PYTHON_CLASSIC)
#  error "This python version requires swig to be run with the '-classic' option"
# endif
#endif
#if (PY_VERSION_HEX <= 0x02020000)
# error "This python version requires swig to be run with the '-nomodern' option"
#endif
#if (PY_VERSION_HEX <= 0x02020000)
# error "This python version requires swig to be run with the '-nomodernargs' option"
#endif

/*-----------------------------------------------
              @(target):= _delta.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__delta

#else
#  define SWIG_init    init_delta

#endif
#define SWIG_name    "_delta"

#define SWIGVERSION 0x030012 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_py.h"
#include "swigutil_py3c.h"


static PyObject * _global_py_pool = NULL;


#define svn_argnum_obj0 1
#define svn_argnum_obj1 2
#define svn_argnum_obj2 3
#define svn_argnum_obj3 4
#define svn_argnum_obj4 5
#define svn_argnum_obj5 6
#define svn_argnum_obj6 7
#define svn_argnum_obj7 8
#define svn_argnum_obj8 9
#define svn_argnum_obj9 10
#define svn_argnum_obj10 11
#define svn_argnum_obj11 12
#define svn_argnum_obj12 13
#define svn_argnum_obj13 14
#define svn_argnum_obj14 15
#define svn_argnum_obj15 16
#define svn_argnum_obj16 17
#define svn_argnum_obj17 18
#define svn_argnum_obj18 19
#define svn_argnum_obj19 20
#define svn_argnum_obj20 21
#define svn_argnum_obj21 22
#define svn_argnum_obj22 23
#define svn_argnum_obj23 24
#define svn_argnum_obj24 25
#define svn_argnum_obj25 26
#define svn_argnum_obj26 27
#define svn_argnum_obj27 28
#define svn_argnum_obj28 29
#define svn_argnum_obj29 30
#define svn_argnum_obj30 31
#define svn_argnum_obj31 32
#define svn_argnum_obj32 33
#define svn_argnum_obj33 34
#define svn_argnum_obj34 35
#define svn_argnum_obj35 36
#define svn_argnum_obj36 37
#define svn_argnum_obj37 38
#define svn_argnum_obj38 39
#define svn_argnum_obj39 40


#include <apr_md5.h>
#include "svn_md5.h"


#include "svn_delta.h"


  #define SWIG_From_long   PyInt_FromLong 


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE long
SWIG_As_long (PyObject* obj)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE unsigned long
SWIG_As_unsigned_SS_long (PyObject* obj)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(unsigned long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong((long)(value));
}

SWIGINTERN void svn_txdelta_window_t__ops_get(struct svn_txdelta_window_t *self,int *num_ops,svn_txdelta_op_t const **ops){
  *num_ops = self->num_ops;
  *ops = self->ops;
}

SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
#if PY_VERSION_HEX>=0x03000000
#if !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    if(alloc) *alloc = SWIG_NEWOBJ;
#endif
    PyBytes_AsStringAndSize(obj, &cstr, &len);
#else
    PyString_AsStringAndSize(obj, &cstr, &len);
#endif
    if (cptr) {
      if (alloc) {
	/* 
	   In python the user should not be able to modify the inner
	   string representation. To warranty that, if you define
	   SWIG_PYTHON_SAFE_CSTRINGS, a new/copy of the python string
	   buffer is always returned.

	   The default behavior is just to return the pointer value,
	   so, be careful.
	*/ 
#if defined(SWIG_PYTHON_SAFE_CSTRINGS)
	if (*alloc != SWIG_OLDOBJ) 
#else
	if (*alloc == SWIG_NEWOBJ) 
#endif
	{
	  *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
	  *alloc = SWIG_NEWOBJ;
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
	*cptr = PyBytes_AsString(obj);
#else
	assert(0); /* Should never reach here with Unicode strings in Python 3 */
#endif
#else
	*cptr = SWIG_Python_str_AsChar(obj);
#endif
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000 && !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    Py_XDECREF(obj);
#endif
    return SWIG_OK;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}





static svn_error_t * svn_delta_editor_invoke_set_target_revision(
  svn_delta_editor_t * _obj, void *edit_baton, svn_revnum_t target_revision, apr_pool_t *scratch_pool) {
  return (_obj->set_target_revision)(edit_baton, target_revision, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_open_root(
  svn_delta_editor_t * _obj, void *edit_baton, svn_revnum_t base_revision, apr_pool_t *result_pool, void **root_baton) {
  return (_obj->open_root)(edit_baton, base_revision, result_pool, root_baton);
}

static svn_error_t * svn_delta_editor_invoke_delete_entry(
  svn_delta_editor_t * _obj, const char *path, svn_revnum_t revision, void *parent_baton, apr_pool_t *scratch_pool) {
  return (_obj->delete_entry)(path, revision, parent_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_add_directory(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, const char *copyfrom_path, svn_revnum_t copyfrom_revision, apr_pool_t *result_pool, void **child_baton) {
  return (_obj->add_directory)(path, parent_baton, copyfrom_path, copyfrom_revision, result_pool, child_baton);
}

static svn_error_t * svn_delta_editor_invoke_open_directory(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, svn_revnum_t base_revision, apr_pool_t *result_pool, void **child_baton) {
  return (_obj->open_directory)(path, parent_baton, base_revision, result_pool, child_baton);
}

static svn_error_t * svn_delta_editor_invoke_change_dir_prop(
  svn_delta_editor_t * _obj, void *dir_baton, const char *name, const svn_string_t *value, apr_pool_t *scratch_pool) {
  return (_obj->change_dir_prop)(dir_baton, name, value, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_close_directory(
  svn_delta_editor_t * _obj, void *dir_baton, apr_pool_t *scratch_pool) {
  return (_obj->close_directory)(dir_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_absent_directory(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, apr_pool_t *scratch_pool) {
  return (_obj->absent_directory)(path, parent_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_add_file(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, const char *copyfrom_path, svn_revnum_t copyfrom_revision, apr_pool_t *result_pool, void **file_baton) {
  return (_obj->add_file)(path, parent_baton, copyfrom_path, copyfrom_revision, result_pool, file_baton);
}

static svn_error_t * svn_delta_editor_invoke_open_file(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, svn_revnum_t base_revision, apr_pool_t *result_pool, void **file_baton) {
  return (_obj->open_file)(path, parent_baton, base_revision, result_pool, file_baton);
}

static svn_error_t * svn_delta_editor_invoke_apply_textdelta(
  svn_delta_editor_t * _obj, void *file_baton, const char *base_checksum, apr_pool_t *result_pool, svn_txdelta_window_handler_t *handler, void **handler_baton) {
  return (_obj->apply_textdelta)(file_baton, base_checksum, result_pool, handler, handler_baton);
}

static svn_error_t * svn_delta_editor_invoke_change_file_prop(
  svn_delta_editor_t * _obj, void *file_baton, const char *name, const svn_string_t *value, apr_pool_t *scratch_pool) {
  return (_obj->change_file_prop)(file_baton, name, value, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_close_file(
  svn_delta_editor_t * _obj, void *file_baton, const char *text_checksum, apr_pool_t *scratch_pool) {
  return (_obj->close_file)(file_baton, text_checksum, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_absent_file(
  svn_delta_editor_t * _obj, const char *path, void *parent_baton, apr_pool_t *scratch_pool) {
  return (_obj->absent_file)(path, parent_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_close_edit(
  svn_delta_editor_t * _obj, void *edit_baton, apr_pool_t *scratch_pool) {
  return (_obj->close_edit)(edit_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_abort_edit(
  svn_delta_editor_t * _obj, void *edit_baton, apr_pool_t *scratch_pool) {
  return (_obj->abort_edit)(edit_baton, scratch_pool);
}

static svn_error_t * svn_delta_editor_invoke_apply_textdelta_stream(
  svn_delta_editor_t * _obj, const struct svn_delta_editor_t *editor, void *file_baton, const char *base_checksum, svn_txdelta_stream_open_func_t open_func, void *open_baton, apr_pool_t *scratch_pool) {
  return (_obj->apply_textdelta_stream)(editor, file_baton, base_checksum, open_func, open_baton, scratch_pool);
}

static svn_error_t * svn_txdelta_invoke_window_handler(
  svn_txdelta_window_handler_t _obj, svn_txdelta_window_t *window, void *baton) {
  return _obj(window, baton);
}

static svn_error_t * svn_txdelta_invoke_next_window_fn(
  svn_txdelta_next_window_fn_t _obj, svn_txdelta_window_t **window, void *baton, apr_pool_t *pool) {
  return _obj(window, baton, pool);
}

static const unsigned char * svn_txdelta_invoke_md5_digest_fn(
  svn_txdelta_md5_digest_fn_t _obj, void *baton) {
  return _obj(baton);
}

static svn_error_t * svn_txdelta_invoke_stream_open_func(
  svn_txdelta_stream_open_func_t _obj, svn_txdelta_stream_t **txdelta_stream, void *baton, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(txdelta_stream, baton, result_pool, scratch_pool);
}

static svn_error_t * svn_delta_invoke_path_driver_cb_func2(
  svn_delta_path_driver_cb_func2_t _obj, void **dir_baton, const svn_delta_editor_t *editor, void *edit_baton, void *parent_baton, void *callback_baton, const char *relpath, apr_pool_t *pool) {
  return _obj(dir_baton, editor, edit_baton, parent_baton, callback_baton, relpath, pool);
}

static svn_error_t * svn_delta_invoke_path_driver_cb_func(
  svn_delta_path_driver_cb_func_t _obj, void **dir_baton, void *parent_baton, void *callback_baton, const char *path, apr_pool_t *pool) {
  return _obj(dir_baton, parent_baton, callback_baton, path, pool);
}

static svn_error_t * svn_file_invoke_rev_handler(
  svn_file_rev_handler_t _obj, void *baton, const char *path, svn_revnum_t rev, apr_hash_t *rev_props, svn_boolean_t result_of_merge, svn_txdelta_window_handler_t *delta_handler, void **delta_baton, apr_array_header_t *prop_diffs, apr_pool_t *pool) {
  return _obj(baton, path, rev, rev_props, result_of_merge, delta_handler, delta_baton, prop_diffs, pool);
}

static svn_error_t * svn_file_invoke_rev_handler_old(
  svn_file_rev_handler_old_t _obj, void *baton, const char *path, svn_revnum_t rev, apr_hash_t *rev_props, svn_txdelta_window_handler_t *delta_handler, void **delta_baton, apr_array_header_t *prop_diffs, apr_pool_t *pool) {
  return _obj(baton, path, rev, rev_props, delta_handler, delta_baton, prop_diffs, pool);
}


#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_svn_swig_py_make_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  PyObject * obj0 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_swig_py_make_editor",0,1,&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_swig_py_make_editor((struct svn_delta_editor_t const **)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_version(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_version_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_version",0,0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_version_t *)svn_delta_version();
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_action_code_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  enum svn_delta_action arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_action_code_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (enum svn_delta_action)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->action_code = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_action_code_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  PyObject * obj0 = 0 ;
  enum svn_delta_action result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_action_code_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (enum svn_delta_action) ((arg1)->action_code);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_offset_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  apr_size_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_offset_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_size_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->offset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_offset_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_size_t result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_offset_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->offset);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_length_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  apr_size_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_length_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_size_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->length = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_op_t_length_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_size_t result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_op_t_length_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->length);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_txdelta_op_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"new_svn_txdelta_op_t",0,0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_txdelta_op_t *)calloc(1, sizeof(struct svn_txdelta_op_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_txdelta_op_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_txdelta_op_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_op_t *arg1 = (struct svn_txdelta_op_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"delete_svn_txdelta_op_t",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_op_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_op_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_txdelta_op_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_txdelta_op_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_sview_offset_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  svn_filesize_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_sview_offset_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg2 = (svn_filesize_t) PyLong_AsLongLong(obj1); 
  if (arg1) (arg1)->sview_offset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_sview_offset_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_filesize_t result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_sview_offset_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->sview_offset);
  resultobj = PyLong_FromLongLong((apr_int64_t)(result)); 
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_sview_len_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  apr_size_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_sview_len_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_size_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->sview_len = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_sview_len_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_size_t result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_sview_len_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->sview_len);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_tview_len_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  apr_size_t arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_tview_len_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_size_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->tview_len = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_tview_len_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  apr_size_t result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_tview_len_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result =  ((arg1)->tview_len);
  resultobj = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_num_ops_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  int arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_num_ops_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->num_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_num_ops_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  int result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_num_ops_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (int) ((arg1)->num_ops);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_src_ops_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  int arg2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_src_ops_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->src_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_src_ops_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  int result;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_src_ops_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (int) ((arg1)->src_ops);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t_new_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_string_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t_new_data_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_string_t *) ((arg1)->new_data);
  {
    resultobj = PyBytes_FromStringAndSize(result->data, result->len);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_window_t__ops_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  int *arg2 = (int *) 0 ;
  svn_txdelta_op_t **arg3 = (svn_txdelta_op_t **) 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  svn_txdelta_op_t *temp3 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_t__ops_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_window_t__ops_get(arg1,arg2,(struct svn_txdelta_op_t const **)arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    apr_pool_t *parent_pool;
    PyObject *parent_py_pool;
    PyObject *ops_list;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &parent_py_pool, &parent_pool))
    SWIG_fail;
    
    ops_list = svn_swig_py_convert_txdelta_op_c_array(*arg2, *arg3,
      SWIGTYPE_p_svn_txdelta_op_t, parent_py_pool);
    
    if (!ops_list) SWIG_fail;
    
    resultobj = SWIG_Python_AppendOutput(resultobj, ops_list);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_txdelta_window_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"new_svn_txdelta_window_t",0,0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_txdelta_window_t *)calloc(1, sizeof(struct svn_txdelta_window_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_txdelta_window_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_txdelta_window_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_txdelta_window_t *arg1 = (struct svn_txdelta_window_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"delete_svn_txdelta_window_t",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_txdelta_window_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_txdelta_window_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_txdelta_window_dup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t *arg1 = (svn_txdelta_window_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_txdelta_window_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_window_dup",1,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_txdelta_window_t *)svn_txdelta_window_dup((struct svn_txdelta_window_t const *)arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_txdelta_window_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_compose_windows(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t *arg1 = (svn_txdelta_window_t *) 0 ;
  svn_txdelta_window_t *arg2 = (svn_txdelta_window_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_txdelta_window_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_compose_windows",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_txdelta_window_t *)svn_txdelta_compose_windows((struct svn_txdelta_window_t const *)arg1,(struct svn_txdelta_window_t const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_txdelta_window_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_apply_instructions(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t *arg1 = (svn_txdelta_window_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_size_t *arg4 = (apr_size_t *) 0 ;
  apr_size_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  arg4 = &temp4;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_apply_instructions",3,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_txdelta_apply_instructions", "sbuf");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_txdelta_apply_instructions", "tbuf");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_apply_instructions(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_apr_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_run(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_txdelta_window_handler_t arg3 = (svn_txdelta_window_handler_t) 0 ;
  void *arg4 = (void *) 0 ;
  svn_checksum_kind_t arg5 ;
  svn_checksum_t **arg6 = (svn_checksum_t **) 0 ;
  svn_cancel_func_t arg7 = (svn_cancel_func_t) 0 ;
  void *arg8 = (void *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_checksum_t *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg6 = &temp6;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_run",6,8,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    }
    else {
      arg1 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg1 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    }
    else {
      arg2 = svn_swig_py_make_stream (obj1, _global_pool);
      if (arg2 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj2);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg3 = *tmp;
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_checksum_kind_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg8 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_run(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg6, SWIGTYPE_p_svn_checksum_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_stream_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void *arg1 = (void *) 0 ;
  svn_txdelta_next_window_fn_t arg2 = (svn_txdelta_next_window_fn_t) 0 ;
  svn_txdelta_md5_digest_fn_t arg3 = (svn_txdelta_md5_digest_fn_t) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_txdelta_stream_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_stream_create",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    } else if (SWIG_ConvertPtr(obj0, (void **) &arg1, 0, 0) == -1) {
      arg1 = (void *) obj0;
      PyErr_Clear();
    }
  }
  {
    svn_txdelta_next_window_fn_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  {
    svn_txdelta_md5_digest_fn_t * tmp =
    svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_p_f_p_void__p_unsigned_char, svn_argnum_obj2);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg3 = *tmp;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_txdelta_stream_t *)svn_txdelta_stream_create(arg1,arg2,(unsigned char const *(*)(void *))arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_txdelta_stream_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_next_window(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t **arg1 = (svn_txdelta_window_t **) 0 ;
  svn_txdelta_stream_t *arg2 = (svn_txdelta_stream_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_window_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_next_window",1,2,&obj0,&obj1)) SWIG_fail;
  {
    arg2 = (svn_txdelta_stream_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_stream_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_next_window(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_txdelta_window_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_md5_digest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_t *arg1 = (svn_txdelta_stream_t *) 0 ;
  PyObject * obj0 = 0 ;
  unsigned char *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_md5_digest",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (svn_txdelta_stream_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_stream_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (unsigned char *)svn_txdelta_md5_digest(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_unsigned_char,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_t **arg1 = (svn_txdelta_stream_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_stream_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta2",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg2 = NULL;
    }
    else {
      arg2 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg2 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    }
    else {
      arg3 = svn_swig_py_make_stream (obj1, _global_pool);
      if (arg3 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta2(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_txdelta_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_t **arg1 = (svn_txdelta_stream_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_stream_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg2 = NULL;
    }
    else {
      arg2 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg2 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    }
    else {
      arg3 = svn_swig_py_make_stream (obj1, _global_pool);
      if (arg3 == NULL) {
        SWIG_fail;
      }
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_txdelta_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_target_push(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_handler_t arg1 = (svn_txdelta_window_handler_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_stream_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_target_push",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    }
    else {
      arg3 = svn_swig_py_make_stream (obj2, _global_pool);
      if (arg3 == NULL) {
        SWIG_fail;
      }
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_stream_t *)svn_txdelta_target_push(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_stream_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_send_string(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_string_t *arg1 = (svn_string_t *) 0 ;
  svn_txdelta_window_handler_t arg2 = (svn_txdelta_window_handler_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_send_string",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None)
    arg1 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj0)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj0, (char **)&(value1.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value1.len = pyStrLen;
      arg1 = &value1;
    }
  }
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_send_string((struct svn_string_t const *)arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_send_stream(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_txdelta_window_handler_t arg2 = (svn_txdelta_window_handler_t) 0 ;
  void *arg3 = (void *) 0 ;
  unsigned char *arg4 = (unsigned char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  unsigned char temp4[APR_MD5_DIGESTSIZE] ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg4 = temp4;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_send_stream",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    }
    else {
      arg1 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg1 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_send_stream(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, PyBytes_FromStringAndSize((const char *)arg4, APR_MD5_DIGESTSIZE));
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_send_txstream(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_t *arg1 = (svn_txdelta_stream_t *) 0 ;
  svn_txdelta_window_handler_t arg2 = (svn_txdelta_window_handler_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_send_txstream",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_txdelta_stream_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_stream_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_send_txstream(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_send_contents(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned char *arg1 = (unsigned char *) 0 ;
  apr_size_t arg2 ;
  svn_txdelta_window_handler_t arg3 = (svn_txdelta_window_handler_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_send_contents",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (unsigned char *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_unsigned_char, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_size_t)SWIG_As_unsigned_SS_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj2);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg3 = *tmp;
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_send_contents((unsigned char const *)arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_apply(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  unsigned char *arg3 = (unsigned char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  svn_txdelta_window_handler_t *arg6 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg7 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  unsigned char temp3[APR_MD5_DIGESTSIZE] ;
  void *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg3 = NULL;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg6 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg6 == NULL) SWIG_fail;
  }
  arg7 = &temp7;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_apply",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    }
    else {
      arg1 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg1 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    }
    else {
      arg2 = svn_swig_py_make_stream (obj1, _global_pool);
      if (arg2 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj2, TRUE, "svn_txdelta_apply", "error_info");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_apply(arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg6, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg7, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_to_svndiff3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_handler_t *arg1 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg2 = (void **) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  int arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg1 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg1 == NULL) SWIG_fail;
  }
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_to_svndiff3",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg3 = NULL;
    }
    else {
      arg3 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg3 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    arg4 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_to_svndiff3(arg1,arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg1, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_to_svndiff2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_handler_t *arg1 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg2 = (void **) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg1 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg1 == NULL) SWIG_fail;
  }
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_to_svndiff2",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg3 = NULL;
    }
    else {
      arg3 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg3 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    arg4 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_to_svndiff2(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg1, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_to_svndiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  svn_txdelta_window_handler_t *arg3 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg4 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg3 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg3 == NULL) SWIG_fail;
  }
  arg4 = &temp4;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_to_svndiff",1,2,&obj0,&obj1)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg1 = NULL;
    }
    else {
      arg1 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg1 == NULL) {
        SWIG_fail;
      }
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_txdelta_to_svndiff(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg3, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_to_svndiff_stream(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_t *arg1 = (svn_txdelta_stream_t *) 0 ;
  int arg2 ;
  int arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_stream_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_to_svndiff_stream",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_txdelta_stream_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_stream_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = (int)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_stream_t *)svn_txdelta_to_svndiff_stream(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_stream_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_parse_svndiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_handler_t arg1 = (svn_txdelta_window_handler_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_stream_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_parse_svndiff",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_stream_t *)svn_txdelta_parse_svndiff(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_stream_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_read_svndiff_window(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t **arg1 = (svn_txdelta_window_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  int arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_window_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_read_svndiff_window",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg2 = NULL;
    }
    else {
      arg2 = svn_swig_py_make_stream (obj0, _global_pool);
      if (arg2 == NULL) {
        SWIG_fail;
      }
    }
  }
  {
    arg3 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_read_svndiff_window(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_txdelta_window_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_skip_svndiff_window(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  int arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_skip_svndiff_window",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = svn_swig_py_make_file(obj0, _global_pool);
    if (!arg1) SWIG_fail;
  }
  {
    arg2 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_skip_svndiff_window(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_set_target_revision_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t,apr_pool_t *) = (svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_set_target_revision_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_set_target_revision_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->set_target_revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_set_target_revision_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_set_target_revision_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *)) ((arg1)->set_target_revision);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_root_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t,apr_pool_t *,void **) = (svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_root_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_open_root_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->open_root = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_root_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t,apr_pool_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_root_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **)) ((arg1)->open_root);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_delete_entry_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,svn_revnum_t,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_delete_entry_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_delete_entry_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->delete_entry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_delete_entry_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_delete_entry_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->delete_entry);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_add_directory_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) = (svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_add_directory_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_add_directory_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->add_directory = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_add_directory_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_add_directory_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)) ((arg1)->add_directory);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_directory_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) = (svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_directory_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_open_directory_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->open_directory = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_directory_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_directory_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)) ((arg1)->open_directory);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_change_dir_prop_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_string_t const *,apr_pool_t *) = (svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_change_dir_prop_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_change_dir_prop_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->change_dir_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_change_dir_prop_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_string_t const *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_change_dir_prop_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)) ((arg1)->change_dir_prop);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_directory_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = (svn_error_t *(*)(void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_directory_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_close_directory_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->close_directory = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_directory_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_directory_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->close_directory);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_absent_directory_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_absent_directory_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_absent_directory_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->absent_directory = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_absent_directory_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_absent_directory_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,apr_pool_t *)) ((arg1)->absent_directory);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_add_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) = (svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_add_file_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_add_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->add_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_add_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_add_file_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)) ((arg1)->add_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) = (svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_file_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_open_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->open_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_open_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_open_file_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)) ((arg1)->open_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_apply_textdelta_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **) = (svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_apply_textdelta_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_apply_textdelta_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **)""'"); 
    }
  }
  if (arg1) (arg1)->apply_textdelta = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_apply_textdelta_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_apply_textdelta_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **)) ((arg1)->apply_textdelta);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_change_file_prop_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_string_t const *,apr_pool_t *) = (svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_change_file_prop_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_change_file_prop_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->change_file_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_change_file_prop_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_string_t const *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_change_file_prop_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)) ((arg1)->change_file_prop);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,char const *,apr_pool_t *) = (svn_error_t *(*)(void *,char const *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_file_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_close_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->close_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_file_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,apr_pool_t *)) ((arg1)->close_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_absent_file_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(char const *,void *,apr_pool_t *) = (svn_error_t *(*)(char const *,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_absent_file_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_absent_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(char const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->absent_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_absent_file_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(char const *,void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_absent_file_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(char const *,void *,apr_pool_t *)) ((arg1)->absent_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_edit_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = (svn_error_t *(*)(void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_edit_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_close_edit_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->close_edit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_close_edit_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_close_edit_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->close_edit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_abort_edit_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = (svn_error_t *(*)(void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_abort_edit_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_abort_edit_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->abort_edit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_abort_edit_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_abort_edit_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->abort_edit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_apply_textdelta_stream_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  svn_error_t *(*arg2)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *) = (svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *)) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_apply_textdelta_stream_set",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_delta_editor_t_apply_textdelta_stream_set" "', argument " "2"" of type '" "svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->apply_textdelta_stream = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_t_apply_textdelta_stream_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *) = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_t_apply_textdelta_stream_get",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *)) ((arg1)->apply_textdelta_stream);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_delta_editor_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"new_svn_delta_editor_t",0,0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_delta_editor_t *)calloc(1, sizeof(struct svn_delta_editor_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_delta_editor_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_delta_editor_t(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_delta_editor_t *arg1 = (struct svn_delta_editor_t *) 0 ;
  PyObject * obj0 = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"delete_svn_delta_editor_t",1,1,&obj0)) SWIG_fail;
  {
    arg1 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_delta_editor_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_delta_editor_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_delta_default_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  svn_delta_editor_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_default_editor",0,1,&obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj0);
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_delta_editor_t *)svn_delta_default_editor(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_delta_editor_t,
    _global_py_pool, args);
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_noop_window_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_t *arg1 = (svn_txdelta_window_t *) 0 ;
  void *arg2 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_noop_window_handler",2,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_noop_window_handler(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_get_cancellation_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_cancel_func_t arg1 = (svn_cancel_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_delta_editor_t *arg3 = (svn_delta_editor_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_delta_editor_t **arg5 = (svn_delta_editor_t **) 0 ;
  void **arg6 = (void **) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp5 ;
  void *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg5 = &temp5;
  arg6 = &temp6;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_get_cancellation_editor",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg2 = obj0;
  }
  {
    arg3 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_get_cancellation_editor(arg1,arg2,(struct svn_delta_editor_t const *)arg3,arg4,(struct svn_delta_editor_t const **)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg5, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg6, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_depth_filter_editor(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t **arg1 = (svn_delta_editor_t **) 0 ;
  void **arg2 = (void **) 0 ;
  svn_delta_editor_t *arg3 = (svn_delta_editor_t *) 0 ;
  void *arg4 = (void *) 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_depth_filter_editor",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg3 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_depth_filter_editor((struct svn_delta_editor_t const **)arg1,arg2,(struct svn_delta_editor_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_delta_path_driver_cb_func2_t arg5 = (svn_delta_path_driver_cb_func2_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver3",6,7,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    svn_delta_path_driver_cb_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj4);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg5 = *tmp;
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver3((struct svn_delta_editor_t const *)arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_array_header_t *arg3 = (apr_array_header_t *) 0 ;
  svn_boolean_t arg4 ;
  svn_delta_path_driver_cb_func_t arg5 = (svn_delta_path_driver_cb_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver2",5,6,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj2,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_delta_path_driver_cb_func_t) svn_swig_py_delta_path_driver_cb_func;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver2((struct svn_delta_editor_t const *)arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_revnum_t arg3 ;
  apr_array_header_t *arg4 = (apr_array_header_t *) 0 ;
  svn_delta_path_driver_cb_func_t arg5 = (svn_delta_path_driver_cb_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver",5,6,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj3,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg5 = (svn_delta_path_driver_cb_func_t) svn_swig_py_delta_path_driver_cb_func;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver((struct svn_delta_editor_t const *)arg1,arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver_start(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_path_driver_state_t **arg1 = (svn_delta_path_driver_state_t **) 0 ;
  svn_delta_editor_t *arg2 = (svn_delta_editor_t *) 0 ;
  void *arg3 = (void *) 0 ;
  svn_delta_path_driver_cb_func2_t arg4 = (svn_delta_path_driver_cb_func2_t) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_path_driver_state_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver_start",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg2 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    svn_delta_path_driver_cb_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj2);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg4 = *tmp;
  }
  {
    if (obj3 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver_start(arg1,(struct svn_delta_editor_t const *)arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    /* FIXME: Missing argout typemap: svn_delta_path_driver_start arg 1 (svn_delta_path_driver_state_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_delta_path_driver_start is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver_step(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_path_driver_state_t *arg1 = (svn_delta_path_driver_state_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver_step",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_delta_path_driver_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_path_driver_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_path_driver_step", "relpath");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver_step(arg1,(char const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_path_driver_finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_path_driver_state_t *arg1 = (svn_delta_path_driver_state_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_path_driver_finish",1,2,&obj0,&obj1)) SWIG_fail;
  {
    arg1 = (svn_delta_path_driver_state_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_path_driver_state_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj1);
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_path_driver_finish(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_compat_wrap_file_rev_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_file_rev_handler_t *arg1 = (svn_file_rev_handler_t *) 0 ;
  void **arg2 = (void **) 0 ;
  svn_file_rev_handler_old_t arg3 = (svn_file_rev_handler_old_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg1 = (svn_file_rev_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_file_rev_handler_t));
    if (arg1 == NULL) SWIG_fail;
  }
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_compat_wrap_file_rev_handler",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    svn_file_rev_handler_old_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg3 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_compat_wrap_file_rev_handler(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg1, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_txdelta_stream_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_txdelta_stream_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_delta_path_driver_state_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_delta_path_driver_state_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_set_target_revision(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_revnum_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_set_target_revision",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_set_target_revision(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_open_root(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_revnum_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  void **arg5 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg5 = &temp5;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_open_root",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_open_root(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg5, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_delete_entry(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_revnum_t arg3 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_delete_entry",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_delete_entry", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_delete_entry(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_add_directory(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  void **arg7 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg7 = &temp7;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_add_directory",5,6,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_add_directory", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, TRUE, "svn_delta_editor_invoke_add_directory", "copyfrom_path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_add_directory(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg7, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_open_directory(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  void **arg6 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg6 = &temp6;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_open_directory",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_open_directory", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_open_directory(arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg6, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_change_dir_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_change_dir_prop",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_delta_editor_invoke_change_dir_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj3 == Py_None)
    arg4 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj3)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj3, (char **)&(value4.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value4.len = pyStrLen;
      arg4 = &value4;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_change_dir_prop(arg1,arg2,(char const *)arg3,(struct svn_string_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_close_directory(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_close_directory",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_close_directory(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_absent_directory(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_absent_directory",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_absent_directory", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_absent_directory(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_add_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_revnum_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  void **arg7 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg7 = &temp7;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_add_file",5,6,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_add_file", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, TRUE, "svn_delta_editor_invoke_add_file", "copyfrom_path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj5);
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_add_file(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg7, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_open_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  void **arg6 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg6 = &temp6;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_open_file",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_open_file", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_open_file(arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg6, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_apply_textdelta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  svn_txdelta_window_handler_t *arg5 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg6 = (void **) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg5 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg5 == NULL) SWIG_fail;
  }
  arg6 = &temp6;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_apply_textdelta",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, TRUE, "svn_delta_editor_invoke_apply_textdelta", "base_checksum");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_apply_textdelta(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg5, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg6, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_change_file_prop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_change_file_prop",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_delta_editor_invoke_change_file_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj3 == Py_None)
    arg4 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj3)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj3, (char **)&(value4.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value4.len = pyStrLen;
      arg4 = &value4;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_change_file_prop(arg1,arg2,(char const *)arg3,(struct svn_string_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_close_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_close_file",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, TRUE, "svn_delta_editor_invoke_close_file", "text_checksum");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_close_file(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_absent_file(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_absent_file",3,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_delta_editor_invoke_absent_file", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_absent_file(arg1,(char const *)arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_close_edit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_close_edit",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_close_edit(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_abort_edit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_abort_edit",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_abort_edit(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_editor_invoke_apply_textdelta_stream(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_editor_t *arg1 = (svn_delta_editor_t *) 0 ;
  struct svn_delta_editor_t *arg2 = (struct svn_delta_editor_t *) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_txdelta_stream_open_func_t arg5 = (svn_txdelta_stream_open_func_t) 0 ;
  void *arg6 = (void *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_editor_invoke_apply_textdelta_stream",6,7,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    arg1 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (struct svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, TRUE, "svn_delta_editor_invoke_apply_textdelta_stream", "base_checksum");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_txdelta_stream_open_func_t * tmp =
    svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj4);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg5 = *tmp;
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_editor_invoke_apply_textdelta_stream(arg1,(struct svn_delta_editor_t const *)arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_invoke_window_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_window_handler_t arg1 = (svn_txdelta_window_handler_t) 0 ;
  svn_txdelta_window_t *arg2 = (svn_txdelta_window_t *) 0 ;
  void *arg3 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_invoke_window_handler",3,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    svn_txdelta_window_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg2 = (svn_txdelta_window_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_txdelta_window_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_invoke_window_handler(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_invoke_next_window_fn(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_next_window_fn_t arg1 = (svn_txdelta_next_window_fn_t) 0 ;
  svn_txdelta_window_t **arg2 = (svn_txdelta_window_t **) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_window_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_invoke_next_window_fn",2,3,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    svn_txdelta_next_window_fn_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_invoke_next_window_fn(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_txdelta_window_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_invoke_md5_digest_fn(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_md5_digest_fn_t arg1 = (svn_txdelta_md5_digest_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  unsigned char *result = 0 ;
  
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_invoke_md5_digest_fn",2,2,&obj0,&obj1)) SWIG_fail;
  {
    svn_txdelta_md5_digest_fn_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void__p_unsigned_char, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (unsigned char *)svn_txdelta_invoke_md5_digest_fn((unsigned char const *(*)(void *))arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_unsigned_char,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_txdelta_invoke_stream_open_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_txdelta_stream_open_func_t arg1 = (svn_txdelta_stream_open_func_t) 0 ;
  svn_txdelta_stream_t **arg2 = (svn_txdelta_stream_t **) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_txdelta_stream_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_txdelta_invoke_stream_open_func",2,4,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    svn_txdelta_stream_open_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj2);
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj3);
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_txdelta_invoke_stream_open_func(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_txdelta_stream_t,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_invoke_path_driver_cb_func2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_path_driver_cb_func2_t arg1 = (svn_delta_path_driver_cb_func2_t) 0 ;
  void **arg2 = (void **) 0 ;
  svn_delta_editor_t *arg3 = (svn_delta_editor_t *) 0 ;
  void *arg4 = (void *) 0 ;
  void *arg5 = (void *) 0 ;
  void *arg6 = (void *) 0 ;
  char *arg7 = (char *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_invoke_path_driver_cb_func2",6,7,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    svn_delta_path_driver_cb_func2_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg3 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    if (obj3 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    if (obj4 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj5, FALSE, "svn_delta_invoke_path_driver_cb_func2", "relpath");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_invoke_path_driver_cb_func2(arg1,arg2,(struct svn_delta_editor_t const *)arg3,arg4,arg5,arg6,(char const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_delta_invoke_path_driver_cb_func(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_delta_path_driver_cb_func_t arg1 = (svn_delta_path_driver_cb_func_t) 0 ;
  void **arg2 = (void **) 0 ;
  void *arg3 = (void *) 0 ;
  void *arg4 = (void *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg2 = &temp2;
  if(!PyArg_UnpackTuple(args,(char *)"svn_delta_invoke_path_driver_cb_func",4,5,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    svn_delta_path_driver_cb_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_delta_invoke_path_driver_cb_func", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj4);
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_delta_invoke_path_driver_cb_func(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_file_invoke_rev_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_file_rev_handler_t arg1 = (svn_file_rev_handler_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  svn_boolean_t arg6 ;
  svn_txdelta_window_handler_t *arg7 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg8 = (void **) 0 ;
  apr_array_header_t *arg9 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp8 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg7 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg7 == NULL) SWIG_fail;
  }
  arg8 = &temp8;
  if(!PyArg_UnpackTuple(args,(char *)"svn_file_invoke_rev_handler",7,8,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  {
    svn_file_rev_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_file_invoke_rev_handler", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_hash_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj7);
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_file_invoke_rev_handler(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg7, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg8, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_file_invoke_rev_handler_old(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  svn_file_rev_handler_old_t arg1 = (svn_file_rev_handler_old_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  svn_txdelta_window_handler_t *arg6 = (svn_txdelta_window_handler_t *) 0 ;
  void **arg7 = (void **) 0 ;
  apr_array_header_t *arg8 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg9 = (apr_pool_t *) 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg6 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg6 == NULL) SWIG_fail;
  }
  arg7 = &temp7;
  if(!PyArg_UnpackTuple(args,(char *)"svn_file_invoke_rev_handler_old",6,7,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  {
    svn_file_rev_handler_old_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_file_invoke_rev_handler_old", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_hash_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(SWIGTYPE_p_apr_pool_t), obj6);
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_file_invoke_rev_handler_old(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg6, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args))
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg7, SWIGTYPE_p_void,
        _global_py_pool, args))
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_txdelta_window_handler_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_txdelta_next_window_fn_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_txdelta_md5_digest_fn_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void__p_unsigned_char, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_txdelta_stream_open_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_delta_path_driver_cb_func2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_delta_path_driver_cb_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_file_rev_handler_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_file_rev_handler_old_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_UnpackTuple(args,(char *)"swigregister", 1, 1,&obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

static PyMethodDef SwigMethods[] = {
	 { (char *)"SWIG_PyInstanceMethod_New", (PyCFunction)SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { (char *)"svn_swig_py_make_editor", _wrap_svn_swig_py_make_editor, METH_VARARGS, (char *)"svn_swig_py_make_editor(apr_pool_t pool)"},
	 { (char *)"svn_delta_version", _wrap_svn_delta_version, METH_VARARGS, (char *)"svn_delta_version() -> svn_version_t const *"},
	 { (char *)"svn_txdelta_op_t_action_code_set", _wrap_svn_txdelta_op_t_action_code_set, METH_VARARGS, (char *)"svn_txdelta_op_t_action_code_set(svn_txdelta_op_t self, enum svn_delta_action action_code)"},
	 { (char *)"svn_txdelta_op_t_action_code_get", _wrap_svn_txdelta_op_t_action_code_get, METH_VARARGS, (char *)"svn_txdelta_op_t_action_code_get(svn_txdelta_op_t self) -> enum svn_delta_action"},
	 { (char *)"svn_txdelta_op_t_offset_set", _wrap_svn_txdelta_op_t_offset_set, METH_VARARGS, (char *)"svn_txdelta_op_t_offset_set(svn_txdelta_op_t self, apr_size_t offset)"},
	 { (char *)"svn_txdelta_op_t_offset_get", _wrap_svn_txdelta_op_t_offset_get, METH_VARARGS, (char *)"svn_txdelta_op_t_offset_get(svn_txdelta_op_t self) -> apr_size_t"},
	 { (char *)"svn_txdelta_op_t_length_set", _wrap_svn_txdelta_op_t_length_set, METH_VARARGS, (char *)"svn_txdelta_op_t_length_set(svn_txdelta_op_t self, apr_size_t length)"},
	 { (char *)"svn_txdelta_op_t_length_get", _wrap_svn_txdelta_op_t_length_get, METH_VARARGS, (char *)"svn_txdelta_op_t_length_get(svn_txdelta_op_t self) -> apr_size_t"},
	 { (char *)"new_svn_txdelta_op_t", _wrap_new_svn_txdelta_op_t, METH_VARARGS, (char *)"new_svn_txdelta_op_t() -> svn_txdelta_op_t"},
	 { (char *)"delete_svn_txdelta_op_t", _wrap_delete_svn_txdelta_op_t, METH_VARARGS, (char *)"delete_svn_txdelta_op_t(svn_txdelta_op_t self)"},
	 { (char *)"svn_txdelta_op_t_swigregister", svn_txdelta_op_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_txdelta_window_t_sview_offset_set", _wrap_svn_txdelta_window_t_sview_offset_set, METH_VARARGS, (char *)"svn_txdelta_window_t_sview_offset_set(svn_txdelta_window_t self, svn_filesize_t sview_offset)"},
	 { (char *)"svn_txdelta_window_t_sview_offset_get", _wrap_svn_txdelta_window_t_sview_offset_get, METH_VARARGS, (char *)"svn_txdelta_window_t_sview_offset_get(svn_txdelta_window_t self) -> svn_filesize_t"},
	 { (char *)"svn_txdelta_window_t_sview_len_set", _wrap_svn_txdelta_window_t_sview_len_set, METH_VARARGS, (char *)"svn_txdelta_window_t_sview_len_set(svn_txdelta_window_t self, apr_size_t sview_len)"},
	 { (char *)"svn_txdelta_window_t_sview_len_get", _wrap_svn_txdelta_window_t_sview_len_get, METH_VARARGS, (char *)"svn_txdelta_window_t_sview_len_get(svn_txdelta_window_t self) -> apr_size_t"},
	 { (char *)"svn_txdelta_window_t_tview_len_set", _wrap_svn_txdelta_window_t_tview_len_set, METH_VARARGS, (char *)"svn_txdelta_window_t_tview_len_set(svn_txdelta_window_t self, apr_size_t tview_len)"},
	 { (char *)"svn_txdelta_window_t_tview_len_get", _wrap_svn_txdelta_window_t_tview_len_get, METH_VARARGS, (char *)"svn_txdelta_window_t_tview_len_get(svn_txdelta_window_t self) -> apr_size_t"},
	 { (char *)"svn_txdelta_window_t_num_ops_set", _wrap_svn_txdelta_window_t_num_ops_set, METH_VARARGS, (char *)"svn_txdelta_window_t_num_ops_set(svn_txdelta_window_t self, int num_ops)"},
	 { (char *)"svn_txdelta_window_t_num_ops_get", _wrap_svn_txdelta_window_t_num_ops_get, METH_VARARGS, (char *)"svn_txdelta_window_t_num_ops_get(svn_txdelta_window_t self) -> int"},
	 { (char *)"svn_txdelta_window_t_src_ops_set", _wrap_svn_txdelta_window_t_src_ops_set, METH_VARARGS, (char *)"svn_txdelta_window_t_src_ops_set(svn_txdelta_window_t self, int src_ops)"},
	 { (char *)"svn_txdelta_window_t_src_ops_get", _wrap_svn_txdelta_window_t_src_ops_get, METH_VARARGS, (char *)"svn_txdelta_window_t_src_ops_get(svn_txdelta_window_t self) -> int"},
	 { (char *)"svn_txdelta_window_t_new_data_get", _wrap_svn_txdelta_window_t_new_data_get, METH_VARARGS, (char *)"svn_txdelta_window_t_new_data_get(svn_txdelta_window_t self) -> svn_string_t const *"},
	 { (char *)"svn_txdelta_window_t__ops_get", _wrap_svn_txdelta_window_t__ops_get, METH_VARARGS, (char *)"svn_txdelta_window_t__ops_get(svn_txdelta_window_t self)"},
	 { (char *)"new_svn_txdelta_window_t", _wrap_new_svn_txdelta_window_t, METH_VARARGS, (char *)"new_svn_txdelta_window_t() -> svn_txdelta_window_t"},
	 { (char *)"delete_svn_txdelta_window_t", _wrap_delete_svn_txdelta_window_t, METH_VARARGS, (char *)"delete_svn_txdelta_window_t(svn_txdelta_window_t self)"},
	 { (char *)"svn_txdelta_window_t_swigregister", svn_txdelta_window_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_txdelta_window_dup", _wrap_svn_txdelta_window_dup, METH_VARARGS, (char *)"svn_txdelta_window_dup(svn_txdelta_window_t window, apr_pool_t pool) -> svn_txdelta_window_t"},
	 { (char *)"svn_txdelta_compose_windows", _wrap_svn_txdelta_compose_windows, METH_VARARGS, (char *)"svn_txdelta_compose_windows(svn_txdelta_window_t window_A, svn_txdelta_window_t window_B, apr_pool_t pool) -> svn_txdelta_window_t"},
	 { (char *)"svn_txdelta_apply_instructions", _wrap_svn_txdelta_apply_instructions, METH_VARARGS, (char *)"svn_txdelta_apply_instructions(svn_txdelta_window_t window, char const * sbuf, char * tbuf)"},
	 { (char *)"svn_txdelta_run", _wrap_svn_txdelta_run, METH_VARARGS, (char *)"svn_txdelta_run(svn_stream_t * source, svn_stream_t * target, svn_txdelta_window_handler_t handler, void * handler_baton, svn_checksum_kind_t checksum_kind, svn_cancel_func_t cancel_func, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_stream_create", _wrap_svn_txdelta_stream_create, METH_VARARGS, (char *)"svn_txdelta_stream_create(void * baton, svn_txdelta_next_window_fn_t next_window, svn_txdelta_md5_digest_fn_t md5_digest, apr_pool_t pool) -> svn_txdelta_stream_t *"},
	 { (char *)"svn_txdelta_next_window", _wrap_svn_txdelta_next_window, METH_VARARGS, (char *)"svn_txdelta_next_window(svn_txdelta_stream_t * stream, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_md5_digest", _wrap_svn_txdelta_md5_digest, METH_VARARGS, (char *)"svn_txdelta_md5_digest(svn_txdelta_stream_t * stream) -> unsigned char const *"},
	 { (char *)"svn_txdelta2", _wrap_svn_txdelta2, METH_VARARGS, (char *)"svn_txdelta2(svn_stream_t * source, svn_stream_t * target, svn_boolean_t calculate_checksum, apr_pool_t pool)"},
	 { (char *)"svn_txdelta", _wrap_svn_txdelta, METH_VARARGS, (char *)"svn_txdelta(svn_stream_t * source, svn_stream_t * target, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_target_push", _wrap_svn_txdelta_target_push, METH_VARARGS, (char *)"svn_txdelta_target_push(svn_txdelta_window_handler_t handler, void * handler_baton, svn_stream_t * source, apr_pool_t pool) -> svn_stream_t *"},
	 { (char *)"svn_txdelta_send_string", _wrap_svn_txdelta_send_string, METH_VARARGS, (char *)"svn_txdelta_send_string(svn_string_t const * string, svn_txdelta_window_handler_t handler, void * handler_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_send_stream", _wrap_svn_txdelta_send_stream, METH_VARARGS, (char *)"svn_txdelta_send_stream(svn_stream_t * stream, svn_txdelta_window_handler_t handler, void * handler_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_send_txstream", _wrap_svn_txdelta_send_txstream, METH_VARARGS, (char *)"svn_txdelta_send_txstream(svn_txdelta_stream_t * txstream, svn_txdelta_window_handler_t handler, void * handler_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_send_contents", _wrap_svn_txdelta_send_contents, METH_VARARGS, (char *)"svn_txdelta_send_contents(unsigned char const * contents, apr_size_t len, svn_txdelta_window_handler_t handler, void * handler_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_apply", _wrap_svn_txdelta_apply, METH_VARARGS, (char *)"svn_txdelta_apply(svn_stream_t * source, svn_stream_t * target, char const * error_info, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_to_svndiff3", _wrap_svn_txdelta_to_svndiff3, METH_VARARGS, (char *)"svn_txdelta_to_svndiff3(svn_stream_t * output, int svndiff_version, int compression_level, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_to_svndiff2", _wrap_svn_txdelta_to_svndiff2, METH_VARARGS, (char *)"svn_txdelta_to_svndiff2(svn_stream_t * output, int svndiff_version, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_to_svndiff", _wrap_svn_txdelta_to_svndiff, METH_VARARGS, (char *)"svn_txdelta_to_svndiff(svn_stream_t * output, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_to_svndiff_stream", _wrap_svn_txdelta_to_svndiff_stream, METH_VARARGS, (char *)"svn_txdelta_to_svndiff_stream(svn_txdelta_stream_t * txstream, int svndiff_version, int compression_level, apr_pool_t pool) -> svn_stream_t *"},
	 { (char *)"svn_txdelta_parse_svndiff", _wrap_svn_txdelta_parse_svndiff, METH_VARARGS, (char *)"svn_txdelta_parse_svndiff(svn_txdelta_window_handler_t handler, void * handler_baton, svn_boolean_t error_on_early_close, apr_pool_t pool) -> svn_stream_t *"},
	 { (char *)"svn_txdelta_read_svndiff_window", _wrap_svn_txdelta_read_svndiff_window, METH_VARARGS, (char *)"svn_txdelta_read_svndiff_window(svn_stream_t * stream, int svndiff_version, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_skip_svndiff_window", _wrap_svn_txdelta_skip_svndiff_window, METH_VARARGS, (char *)"svn_txdelta_skip_svndiff_window(apr_file_t file, int svndiff_version, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_t_set_target_revision_set", _wrap_svn_delta_editor_t_set_target_revision_set, METH_VARARGS, (char *)"svn_delta_editor_t_set_target_revision_set(svn_delta_editor_t self, svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *) set_target_revision)"},
	 { (char *)"svn_delta_editor_t_set_target_revision_get", _wrap_svn_delta_editor_t_set_target_revision_get, METH_VARARGS, (char *)"svn_delta_editor_t_set_target_revision_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_open_root_set", _wrap_svn_delta_editor_t_open_root_set, METH_VARARGS, (char *)"svn_delta_editor_t_open_root_set(svn_delta_editor_t self, svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **) open_root)"},
	 { (char *)"svn_delta_editor_t_open_root_get", _wrap_svn_delta_editor_t_open_root_get, METH_VARARGS, (char *)"svn_delta_editor_t_open_root_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **)"},
	 { (char *)"svn_delta_editor_t_delete_entry_set", _wrap_svn_delta_editor_t_delete_entry_set, METH_VARARGS, (char *)"svn_delta_editor_t_delete_entry_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *) delete_entry)"},
	 { (char *)"svn_delta_editor_t_delete_entry_get", _wrap_svn_delta_editor_t_delete_entry_get, METH_VARARGS, (char *)"svn_delta_editor_t_delete_entry_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_add_directory_set", _wrap_svn_delta_editor_t_add_directory_set, METH_VARARGS, (char *)"svn_delta_editor_t_add_directory_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) add_directory)"},
	 { (char *)"svn_delta_editor_t_add_directory_get", _wrap_svn_delta_editor_t_add_directory_get, METH_VARARGS, (char *)"svn_delta_editor_t_add_directory_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)"},
	 { (char *)"svn_delta_editor_t_open_directory_set", _wrap_svn_delta_editor_t_open_directory_set, METH_VARARGS, (char *)"svn_delta_editor_t_open_directory_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) open_directory)"},
	 { (char *)"svn_delta_editor_t_open_directory_get", _wrap_svn_delta_editor_t_open_directory_get, METH_VARARGS, (char *)"svn_delta_editor_t_open_directory_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)"},
	 { (char *)"svn_delta_editor_t_change_dir_prop_set", _wrap_svn_delta_editor_t_change_dir_prop_set, METH_VARARGS, (char *)"svn_delta_editor_t_change_dir_prop_set(svn_delta_editor_t self, svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *) change_dir_prop)"},
	 { (char *)"svn_delta_editor_t_change_dir_prop_get", _wrap_svn_delta_editor_t_change_dir_prop_get, METH_VARARGS, (char *)"svn_delta_editor_t_change_dir_prop_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_close_directory_set", _wrap_svn_delta_editor_t_close_directory_set, METH_VARARGS, (char *)"svn_delta_editor_t_close_directory_set(svn_delta_editor_t self, svn_error_t *(*)(void *,apr_pool_t *) close_directory)"},
	 { (char *)"svn_delta_editor_t_close_directory_get", _wrap_svn_delta_editor_t_close_directory_get, METH_VARARGS, (char *)"svn_delta_editor_t_close_directory_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_absent_directory_set", _wrap_svn_delta_editor_t_absent_directory_set, METH_VARARGS, (char *)"svn_delta_editor_t_absent_directory_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,apr_pool_t *) absent_directory)"},
	 { (char *)"svn_delta_editor_t_absent_directory_get", _wrap_svn_delta_editor_t_absent_directory_get, METH_VARARGS, (char *)"svn_delta_editor_t_absent_directory_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_add_file_set", _wrap_svn_delta_editor_t_add_file_set, METH_VARARGS, (char *)"svn_delta_editor_t_add_file_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **) add_file)"},
	 { (char *)"svn_delta_editor_t_add_file_get", _wrap_svn_delta_editor_t_add_file_get, METH_VARARGS, (char *)"svn_delta_editor_t_add_file_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)"},
	 { (char *)"svn_delta_editor_t_open_file_set", _wrap_svn_delta_editor_t_open_file_set, METH_VARARGS, (char *)"svn_delta_editor_t_open_file_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **) open_file)"},
	 { (char *)"svn_delta_editor_t_open_file_get", _wrap_svn_delta_editor_t_open_file_get, METH_VARARGS, (char *)"svn_delta_editor_t_open_file_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)"},
	 { (char *)"svn_delta_editor_t_apply_textdelta_set", _wrap_svn_delta_editor_t_apply_textdelta_set, METH_VARARGS, (char *)"svn_delta_editor_t_apply_textdelta_set(svn_delta_editor_t self, svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **) apply_textdelta)"},
	 { (char *)"svn_delta_editor_t_apply_textdelta_get", _wrap_svn_delta_editor_t_apply_textdelta_get, METH_VARARGS, (char *)"svn_delta_editor_t_apply_textdelta_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **)"},
	 { (char *)"svn_delta_editor_t_change_file_prop_set", _wrap_svn_delta_editor_t_change_file_prop_set, METH_VARARGS, (char *)"svn_delta_editor_t_change_file_prop_set(svn_delta_editor_t self, svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *) change_file_prop)"},
	 { (char *)"svn_delta_editor_t_change_file_prop_get", _wrap_svn_delta_editor_t_change_file_prop_get, METH_VARARGS, (char *)"svn_delta_editor_t_change_file_prop_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_close_file_set", _wrap_svn_delta_editor_t_close_file_set, METH_VARARGS, (char *)"svn_delta_editor_t_close_file_set(svn_delta_editor_t self, svn_error_t *(*)(void *,char const *,apr_pool_t *) close_file)"},
	 { (char *)"svn_delta_editor_t_close_file_get", _wrap_svn_delta_editor_t_close_file_get, METH_VARARGS, (char *)"svn_delta_editor_t_close_file_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,char const *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_absent_file_set", _wrap_svn_delta_editor_t_absent_file_set, METH_VARARGS, (char *)"svn_delta_editor_t_absent_file_set(svn_delta_editor_t self, svn_error_t *(*)(char const *,void *,apr_pool_t *) absent_file)"},
	 { (char *)"svn_delta_editor_t_absent_file_get", _wrap_svn_delta_editor_t_absent_file_get, METH_VARARGS, (char *)"svn_delta_editor_t_absent_file_get(svn_delta_editor_t self) -> svn_error_t *(*)(char const *,void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_close_edit_set", _wrap_svn_delta_editor_t_close_edit_set, METH_VARARGS, (char *)"svn_delta_editor_t_close_edit_set(svn_delta_editor_t self, svn_error_t *(*)(void *,apr_pool_t *) close_edit)"},
	 { (char *)"svn_delta_editor_t_close_edit_get", _wrap_svn_delta_editor_t_close_edit_get, METH_VARARGS, (char *)"svn_delta_editor_t_close_edit_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_abort_edit_set", _wrap_svn_delta_editor_t_abort_edit_set, METH_VARARGS, (char *)"svn_delta_editor_t_abort_edit_set(svn_delta_editor_t self, svn_error_t *(*)(void *,apr_pool_t *) abort_edit)"},
	 { (char *)"svn_delta_editor_t_abort_edit_get", _wrap_svn_delta_editor_t_abort_edit_get, METH_VARARGS, (char *)"svn_delta_editor_t_abort_edit_get(svn_delta_editor_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { (char *)"svn_delta_editor_t_apply_textdelta_stream_set", _wrap_svn_delta_editor_t_apply_textdelta_stream_set, METH_VARARGS, (char *)"svn_delta_editor_t_apply_textdelta_stream_set(svn_delta_editor_t self, svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *) apply_textdelta_stream)"},
	 { (char *)"svn_delta_editor_t_apply_textdelta_stream_get", _wrap_svn_delta_editor_t_apply_textdelta_stream_get, METH_VARARGS, (char *)"svn_delta_editor_t_apply_textdelta_stream_get(svn_delta_editor_t self) -> svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *)"},
	 { (char *)"new_svn_delta_editor_t", _wrap_new_svn_delta_editor_t, METH_VARARGS, (char *)"new_svn_delta_editor_t() -> svn_delta_editor_t"},
	 { (char *)"delete_svn_delta_editor_t", _wrap_delete_svn_delta_editor_t, METH_VARARGS, (char *)"delete_svn_delta_editor_t(svn_delta_editor_t self)"},
	 { (char *)"svn_delta_editor_t_swigregister", svn_delta_editor_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_delta_default_editor", _wrap_svn_delta_default_editor, METH_VARARGS, (char *)"svn_delta_default_editor(apr_pool_t pool) -> svn_delta_editor_t"},
	 { (char *)"svn_delta_noop_window_handler", _wrap_svn_delta_noop_window_handler, METH_VARARGS, (char *)"svn_delta_noop_window_handler(svn_txdelta_window_t window, void * baton) -> svn_error_t"},
	 { (char *)"svn_delta_get_cancellation_editor", _wrap_svn_delta_get_cancellation_editor, METH_VARARGS, (char *)"svn_delta_get_cancellation_editor(svn_cancel_func_t cancel_func, svn_delta_editor_t wrapped_editor, void * wrapped_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_delta_depth_filter_editor", _wrap_svn_delta_depth_filter_editor, METH_VARARGS, (char *)"svn_delta_depth_filter_editor(svn_delta_editor_t wrapped_editor, void * wrapped_edit_baton, svn_depth_t requested_depth, svn_boolean_t has_target, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver3", _wrap_svn_delta_path_driver3, METH_VARARGS, (char *)"svn_delta_path_driver3(svn_delta_editor_t editor, void * edit_baton, apr_array_header_t relpaths, svn_boolean_t sort_paths, svn_delta_path_driver_cb_func2_t callback_func, void * callback_baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver2", _wrap_svn_delta_path_driver2, METH_VARARGS, (char *)"svn_delta_path_driver2(svn_delta_editor_t editor, void * edit_baton, apr_array_header_t paths, svn_boolean_t sort_paths, svn_delta_path_driver_cb_func_t callback_func, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver", _wrap_svn_delta_path_driver, METH_VARARGS, (char *)"svn_delta_path_driver(svn_delta_editor_t editor, void * edit_baton, svn_revnum_t revision, apr_array_header_t paths, svn_delta_path_driver_cb_func_t callback_func, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver_start", _wrap_svn_delta_path_driver_start, METH_VARARGS, (char *)"svn_delta_path_driver_start(svn_delta_editor_t editor, void * edit_baton, svn_delta_path_driver_cb_func2_t callback_func, void * callback_baton, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver_step", _wrap_svn_delta_path_driver_step, METH_VARARGS, (char *)"svn_delta_path_driver_step(svn_delta_path_driver_state_t * state, char const * relpath, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_path_driver_finish", _wrap_svn_delta_path_driver_finish, METH_VARARGS, (char *)"svn_delta_path_driver_finish(svn_delta_path_driver_state_t * state, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_compat_wrap_file_rev_handler", _wrap_svn_compat_wrap_file_rev_handler, METH_VARARGS, (char *)"svn_compat_wrap_file_rev_handler(svn_file_rev_handler_old_t handler, void * handler_baton, apr_pool_t pool)"},
	 { (char *)"svn_txdelta_stream_t_swigregister", svn_txdelta_stream_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_delta_path_driver_state_t_swigregister", svn_delta_path_driver_state_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_delta_editor_invoke_set_target_revision", _wrap_svn_delta_editor_invoke_set_target_revision, METH_VARARGS, (char *)"svn_delta_editor_invoke_set_target_revision(svn_delta_editor_t _obj, void * edit_baton, svn_revnum_t target_revision, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_open_root", _wrap_svn_delta_editor_invoke_open_root, METH_VARARGS, (char *)"svn_delta_editor_invoke_open_root(svn_delta_editor_t _obj, void * edit_baton, svn_revnum_t base_revision, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_delete_entry", _wrap_svn_delta_editor_invoke_delete_entry, METH_VARARGS, (char *)"svn_delta_editor_invoke_delete_entry(svn_delta_editor_t _obj, char const * path, svn_revnum_t revision, void * parent_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_add_directory", _wrap_svn_delta_editor_invoke_add_directory, METH_VARARGS, (char *)"svn_delta_editor_invoke_add_directory(svn_delta_editor_t _obj, char const * path, void * parent_baton, char const * copyfrom_path, svn_revnum_t copyfrom_revision, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_open_directory", _wrap_svn_delta_editor_invoke_open_directory, METH_VARARGS, (char *)"svn_delta_editor_invoke_open_directory(svn_delta_editor_t _obj, char const * path, void * parent_baton, svn_revnum_t base_revision, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_change_dir_prop", _wrap_svn_delta_editor_invoke_change_dir_prop, METH_VARARGS, (char *)"svn_delta_editor_invoke_change_dir_prop(svn_delta_editor_t _obj, void * dir_baton, char const * name, svn_string_t const * value, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_close_directory", _wrap_svn_delta_editor_invoke_close_directory, METH_VARARGS, (char *)"svn_delta_editor_invoke_close_directory(svn_delta_editor_t _obj, void * dir_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_absent_directory", _wrap_svn_delta_editor_invoke_absent_directory, METH_VARARGS, (char *)"svn_delta_editor_invoke_absent_directory(svn_delta_editor_t _obj, char const * path, void * parent_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_add_file", _wrap_svn_delta_editor_invoke_add_file, METH_VARARGS, (char *)"svn_delta_editor_invoke_add_file(svn_delta_editor_t _obj, char const * path, void * parent_baton, char const * copyfrom_path, svn_revnum_t copyfrom_revision, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_open_file", _wrap_svn_delta_editor_invoke_open_file, METH_VARARGS, (char *)"svn_delta_editor_invoke_open_file(svn_delta_editor_t _obj, char const * path, void * parent_baton, svn_revnum_t base_revision, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_apply_textdelta", _wrap_svn_delta_editor_invoke_apply_textdelta, METH_VARARGS, (char *)"svn_delta_editor_invoke_apply_textdelta(svn_delta_editor_t _obj, void * file_baton, char const * base_checksum, apr_pool_t result_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_change_file_prop", _wrap_svn_delta_editor_invoke_change_file_prop, METH_VARARGS, (char *)"svn_delta_editor_invoke_change_file_prop(svn_delta_editor_t _obj, void * file_baton, char const * name, svn_string_t const * value, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_close_file", _wrap_svn_delta_editor_invoke_close_file, METH_VARARGS, (char *)"svn_delta_editor_invoke_close_file(svn_delta_editor_t _obj, void * file_baton, char const * text_checksum, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_absent_file", _wrap_svn_delta_editor_invoke_absent_file, METH_VARARGS, (char *)"svn_delta_editor_invoke_absent_file(svn_delta_editor_t _obj, char const * path, void * parent_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_close_edit", _wrap_svn_delta_editor_invoke_close_edit, METH_VARARGS, (char *)"svn_delta_editor_invoke_close_edit(svn_delta_editor_t _obj, void * edit_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_abort_edit", _wrap_svn_delta_editor_invoke_abort_edit, METH_VARARGS, (char *)"svn_delta_editor_invoke_abort_edit(svn_delta_editor_t _obj, void * edit_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_editor_invoke_apply_textdelta_stream", _wrap_svn_delta_editor_invoke_apply_textdelta_stream, METH_VARARGS, (char *)"svn_delta_editor_invoke_apply_textdelta_stream(svn_delta_editor_t _obj, svn_delta_editor_t editor, void * file_baton, char const * base_checksum, svn_txdelta_stream_open_func_t open_func, void * open_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_invoke_window_handler", _wrap_svn_txdelta_invoke_window_handler, METH_VARARGS, (char *)"svn_txdelta_invoke_window_handler(svn_txdelta_window_handler_t _obj, svn_txdelta_window_t window, void * baton) -> svn_error_t"},
	 { (char *)"svn_txdelta_invoke_next_window_fn", _wrap_svn_txdelta_invoke_next_window_fn, METH_VARARGS, (char *)"svn_txdelta_invoke_next_window_fn(svn_txdelta_next_window_fn_t _obj, void * baton, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_invoke_md5_digest_fn", _wrap_svn_txdelta_invoke_md5_digest_fn, METH_VARARGS, (char *)"svn_txdelta_invoke_md5_digest_fn(svn_txdelta_md5_digest_fn_t _obj, void * baton) -> unsigned char const *"},
	 { (char *)"svn_txdelta_invoke_stream_open_func", _wrap_svn_txdelta_invoke_stream_open_func, METH_VARARGS, (char *)"svn_txdelta_invoke_stream_open_func(svn_txdelta_stream_open_func_t _obj, void * baton, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { (char *)"svn_delta_invoke_path_driver_cb_func2", _wrap_svn_delta_invoke_path_driver_cb_func2, METH_VARARGS, (char *)"svn_delta_invoke_path_driver_cb_func2(svn_delta_path_driver_cb_func2_t _obj, svn_delta_editor_t editor, void * edit_baton, void * parent_baton, void * callback_baton, char const * relpath, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_delta_invoke_path_driver_cb_func", _wrap_svn_delta_invoke_path_driver_cb_func, METH_VARARGS, (char *)"svn_delta_invoke_path_driver_cb_func(svn_delta_path_driver_cb_func_t _obj, void * parent_baton, void * callback_baton, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_file_invoke_rev_handler", _wrap_svn_file_invoke_rev_handler, METH_VARARGS, (char *)"svn_file_invoke_rev_handler(svn_file_rev_handler_t _obj, void * baton, char const * path, svn_revnum_t rev, apr_hash_t rev_props, svn_boolean_t result_of_merge, apr_array_header_t prop_diffs, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_file_invoke_rev_handler_old", _wrap_svn_file_invoke_rev_handler_old, METH_VARARGS, (char *)"svn_file_invoke_rev_handler_old(svn_file_rev_handler_old_t _obj, void * baton, char const * path, svn_revnum_t rev, apr_hash_t rev_props, apr_array_header_t prop_diffs, apr_pool_t pool) -> svn_error_t"},
	 { (char *)"svn_txdelta_window_handler_t_swigregister", svn_txdelta_window_handler_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_txdelta_next_window_fn_t_swigregister", svn_txdelta_next_window_fn_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_txdelta_md5_digest_fn_t_swigregister", svn_txdelta_md5_digest_fn_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_txdelta_stream_open_func_t_swigregister", svn_txdelta_stream_open_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_delta_path_driver_cb_func2_t_swigregister", svn_delta_path_driver_cb_func2_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_delta_path_driver_cb_func_t_swigregister", svn_delta_path_driver_cb_func_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_file_rev_handler_t_swigregister", svn_file_rev_handler_t_swigregister, METH_VARARGS, NULL},
	 { (char *)"svn_file_rev_handler_old_t_swigregister", svn_file_rev_handler_old_t_swigregister, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "apr_array_header_t *|svn_rangelist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "apr_int32_t *|time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_int64_t *|svn_filesize_t *|apr_time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_size_t = {"_p_apr_size_t", "apr_size_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_txdelta_stream_t **,void *,apr_pool_t *,apr_pool_t *)|svn_txdelta_stream_open_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_txdelta_window_t **,void *,apr_pool_t *)|svn_txdelta_next_window_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_delta_path_driver_cb_func2_t|svn_error_t *(*)(void **,svn_delta_editor_t const *,void *,void *,void *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_delta_path_driver_cb_func_t|svn_error_t *(*)(void **,void *,void *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(char const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t = {"_p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t", "svn_error_t *(*)(char const *,void *,char const *,svn_revnum_t,apr_pool_t *,void **)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t = {"_p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t", "svn_error_t *(*)(char const *,void *,svn_revnum_t,apr_pool_t *,void **)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(struct svn_delta_editor_t const *,void *,char const *,svn_txdelta_stream_open_func_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t = {"_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_txdelta_window_t *,void *)|svn_txdelta_window_handler_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_unsigned_char = {"_p_f_p_void__p_unsigned_char", "unsigned char *(*)(void *)|svn_txdelta_md5_digest_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t", "svn_error_t *(*)(void *,char const *,apr_pool_t *,svn_txdelta_window_handler_t *,void **)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_string_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_file_rev_handler_old_t|svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_boolean_t,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)|svn_file_rev_handler_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t,apr_pool_t *,void **)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *|svn_boolean_t *|apr_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "long *|svn_revnum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(**)(svn_txdelta_stream_t **,void *,apr_pool_t *,apr_pool_t *)|svn_txdelta_stream_open_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(**)(svn_txdelta_window_t **,void *,apr_pool_t *)|svn_txdelta_next_window_fn_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(**)(void **,svn_delta_editor_t const *,void *,void *,void *,char const *,apr_pool_t *)|svn_delta_path_driver_cb_func2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(**)(void **,void *,void *,char const *,apr_pool_t *)|svn_delta_path_driver_cb_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t = {"_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t", "svn_error_t *(**)(svn_txdelta_window_t *,void *)|svn_txdelta_window_handler_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void__p_unsigned_char = {"_p_p_f_p_void__p_unsigned_char", "svn_txdelta_md5_digest_fn_t *|unsigned char *(**)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_file_rev_handler_old_t *|svn_error_t *(**)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_file_rev_handler_t *|svn_error_t *(**)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_boolean_t,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_checksum_t = {"_p_p_svn_checksum_t", "struct svn_checksum_t **|svn_checksum_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_delta_editor_t = {"_p_p_svn_delta_editor_t", "struct svn_delta_editor_t **|svn_delta_editor_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_delta_path_driver_state_t = {"_p_p_svn_delta_path_driver_state_t", "struct svn_delta_path_driver_state_t **|svn_delta_path_driver_state_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_txdelta_op_t = {"_p_p_svn_txdelta_op_t", "struct svn_txdelta_op_t **|svn_txdelta_op_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_txdelta_stream_t = {"_p_p_svn_txdelta_stream_t", "struct svn_txdelta_stream_t **|svn_txdelta_stream_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_txdelta_window_t = {"_p_p_svn_txdelta_window_t", "struct svn_txdelta_window_t **|svn_txdelta_window_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "struct svn_auth_baton_t *|svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "struct svn_auth_cred_simple_t *|svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "struct svn_auth_cred_ssl_client_cert_pw_t *|svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "struct svn_auth_cred_ssl_client_cert_t *|svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "struct svn_auth_cred_ssl_server_trust_t *|svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "struct svn_auth_cred_username_t *|svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "struct svn_auth_iterstate_t *|svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t *|svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "struct svn_auth_provider_t *|svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "struct svn_auth_ssl_server_cert_info_t *|svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_cache_config_t = {"_p_svn_cache_config_t", "struct svn_cache_config_t *|svn_cache_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "struct svn_checksum_ctx_t *|svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "enum svn_checksum_kind_t *|svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "struct svn_checksum_t *|svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client__shelf_version_t = {"_p_svn_client__shelf_version_t", "svn_client__shelf_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_conflict_option_t = {"_p_svn_client_conflict_option_t", "svn_client_conflict_option_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "struct svn_commit_info_t *|svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "struct svn_config_t *|svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_editor_t = {"_p_svn_delta_editor_t", "struct svn_delta_editor_t *|svn_delta_editor_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_path_driver_state_t = {"_p_svn_delta_path_driver_state_t", "struct svn_delta_path_driver_state_t *|svn_delta_path_driver_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "enum svn_depth_t *|svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "struct svn_dirent_t *|svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "enum svn_errno_t *|svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_fs_lock_target_t = {"_p_svn_fs_lock_target_t", "svn_fs_lock_target_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "struct svn_io_dirent2_t *|svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "struct svn_io_dirent_t *|svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "enum svn_io_file_del_t *|svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "struct svn_location_segment_t *|svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "struct svn_lock_t *|svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "struct svn_log_changed_path2_t *|svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "struct svn_log_changed_path_t *|svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "struct svn_log_entry_t *|svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "struct svn_merge_range_t *|svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "enum svn_mergeinfo_inheritance_t *|svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "enum svn_node_kind_t *|svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "struct svn_opt_revision_range_t *|svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "struct svn_opt_revision_t *|svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "union svn_opt_revision_value_t *|svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "struct svn_opt_subcommand_desc2_t *|svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc3_t = {"_p_svn_opt_subcommand_desc3_t", "struct svn_opt_subcommand_desc3_t *|svn_opt_subcommand_desc3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "struct svn_opt_subcommand_desc_t *|svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "struct svn_prop_inherited_item_t *|svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "struct svn_stream_mark_t *|svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "struct svn_stream_t *|svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "struct svn_string_t *|svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "struct svn_stringbuf_t *|svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "enum svn_tristate_t *|svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_op_t = {"_p_svn_txdelta_op_t", "struct svn_txdelta_op_t *|svn_txdelta_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_stream_t = {"_p_svn_txdelta_stream_t", "struct svn_txdelta_stream_t *|svn_txdelta_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_window_t = {"_p_svn_txdelta_window_t", "struct svn_txdelta_window_t *|svn_txdelta_window_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "struct svn_version_checklist_t *|svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "struct svn_version_ext_linked_lib_t *|svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "struct svn_version_ext_loaded_lib_t *|svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "struct svn_version_extended_t *|svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "struct svn_version_t *|svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item2_t = {"_p_svn_wc_external_item2_t", "svn_wc_external_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "unsigned long *|svn_linenum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_apr_size_t,
  &_swigt__p_char,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void__p_unsigned_char,
  &_swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  &_swigt__p_p_f_p_void__p_unsigned_char,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_svn_checksum_t,
  &_swigt__p_p_svn_delta_editor_t,
  &_swigt__p_p_svn_delta_path_driver_state_t,
  &_swigt__p_p_svn_txdelta_op_t,
  &_swigt__p_p_svn_txdelta_stream_t,
  &_swigt__p_p_svn_txdelta_window_t,
  &_swigt__p_p_void,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_cache_config_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_client__shelf_version_t,
  &_swigt__p_svn_client_conflict_option_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_delta_editor_t,
  &_swigt__p_svn_delta_path_driver_state_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_fs_lock_target_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc3_t,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_txdelta_op_t,
  &_swigt__p_svn_txdelta_stream_t,
  &_swigt__p_svn_txdelta_window_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_svn_wc_external_item2_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_size_t[] = {  {&_swigt__p_apr_size_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_unsigned_char[] = {  {&_swigt__p_f_p_void__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t[] = {  {&_swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void__p_unsigned_char[] = {  {&_swigt__p_p_f_p_void__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_checksum_t[] = {  {&_swigt__p_p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_delta_editor_t[] = {  {&_swigt__p_p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_delta_path_driver_state_t[] = {  {&_swigt__p_p_svn_delta_path_driver_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_txdelta_op_t[] = {  {&_swigt__p_p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_txdelta_stream_t[] = {  {&_swigt__p_p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_txdelta_window_t[] = {  {&_swigt__p_p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_cache_config_t[] = {  {&_swigt__p_svn_cache_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client__shelf_version_t[] = {  {&_swigt__p_svn_client__shelf_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_conflict_option_t[] = {  {&_swigt__p_svn_client_conflict_option_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_editor_t[] = {  {&_swigt__p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_path_driver_state_t[] = {  {&_swigt__p_svn_delta_path_driver_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_fs_lock_target_t[] = {  {&_swigt__p_svn_fs_lock_target_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc3_t[] = {  {&_swigt__p_svn_opt_subcommand_desc3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_op_t[] = {  {&_swigt__p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_stream_t[] = {  {&_swigt__p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_window_t[] = {  {&_swigt__p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item2_t[] = {  {&_swigt__p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_pool_t,
  _swigc__p_apr_size_t,
  _swigc__p_char,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_void_p_q_const__char_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  _swigc__p_f_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__struct_svn_delta_editor_t_p_void_p_q_const__char_svn_txdelta_stream_open_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void__p_unsigned_char,
  _swigc__p_f_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t_p_svn_txdelta_window_handler_t_p_p_void__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_svn_revnum_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_svn_revnum_t_p_apr_pool_t_p_p_void__p_svn_error_t,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_f_p_p_svn_txdelta_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_svn_txdelta_window_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_void_p_q_const__svn_delta_editor_t_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_p_void_p_void_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  _swigc__p_p_f_p_void__p_unsigned_char,
  _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_svn_checksum_t,
  _swigc__p_p_svn_delta_editor_t,
  _swigc__p_p_svn_delta_path_driver_state_t,
  _swigc__p_p_svn_txdelta_op_t,
  _swigc__p_p_svn_txdelta_stream_t,
  _swigc__p_p_svn_txdelta_window_t,
  _swigc__p_p_void,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_cache_config_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_client__shelf_version_t,
  _swigc__p_svn_client_conflict_option_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_delta_editor_t,
  _swigc__p_svn_delta_path_driver_state_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_fs_lock_target_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc3_t,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_txdelta_op_t,
  _swigc__p_svn_txdelta_stream_t,
  _swigc__p_svn_txdelta_window_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_svn_wc_external_item2_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN int
  swig_varlink_print(swig_varlinkobject *v, FILE *fp, int SWIGUNUSEDPARM(flags)) {
    char *tmp;
    PyObject *str = swig_varlink_str(v);
    fprintf(fp,"Swig global variables ");
    fprintf(fp,"%s\n", tmp = SWIG_Python_str_AsChar(str));
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(str);
    return 0;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        (char *)"swigvarlink",              /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        (printfunc) swig_varlink_print,     /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
        0,                                  /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
        0,                                  /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
        0,                                  /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
        0,                                  /* tp_allocs */
        0,                                  /* tp_frees */
        0,                                  /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
        0,                                  /* tp_prev */
#endif
        0                                   /* tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
#if PY_VERSION_HEX < 0x02020000
      varlink_type.ob_type = &PyType_Type;
#else
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
#endif
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        strncpy(gv->name,name,size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *_SWIG_globals = 0; 
    if (!_SWIG_globals) _SWIG_globals = SWIG_newvarlink();  
    return _SWIG_globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              strncpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              strncpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md;
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
# if PY_VERSION_HEX >= 0x03020000
    PyModuleDef_HEAD_INIT,
# else
    {
      PyObject_HEAD_INIT(NULL)
      NULL, /* m_init */
      0,    /* m_index */
      NULL, /* m_copy */
    },
# endif
    (char *) SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    (PyCFunction) SwigPyObject_own,
    (PyCFunction) SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType();
  assert(metatype);
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule((char *) SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "SVN_DELTA_COMPRESSION_LEVEL_NONE",SWIG_From_long((long)(0)));
  SWIG_Python_SetConstant(d, "SVN_DELTA_COMPRESSION_LEVEL_MAX",SWIG_From_long((long)(9)));
  SWIG_Python_SetConstant(d, "SVN_DELTA_COMPRESSION_LEVEL_DEFAULT",SWIG_From_long((long)(5)));
  SWIG_Python_SetConstant(d, "svn_txdelta_source",SWIG_From_long((long)(svn_txdelta_source)));
  SWIG_Python_SetConstant(d, "svn_txdelta_target",SWIG_From_long((long)(svn_txdelta_target)));
  SWIG_Python_SetConstant(d, "svn_txdelta_new",SWIG_From_long((long)(svn_txdelta_new)));
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

